#ifndef CGP_H
#define CGP_H

#include <stdlib.h>
#include <string.h>

namespace CartesianGeneticProgramming {

class RandomNumberSource {
public:

public:
	RandomNumberSource() {
	}

	float Next() {
		return (rand() & 8191) * (1.f/8192.f);
	}

	int NextInt(int range) {
		return rand() % range;
	}
};

class CGPIndividual {
public:
	// nr*nc*[function,parameter,input0,input1,...,inputNA-1], output0, output1,...,outputNO-1
	float *genome;
	int na, ni, no, nr, nc;
	float fitness;
	bool isTested;

	int randomSeed;
	int id;

public:
	CGPIndividual()
		: genome(NULL), na(0), ni(0), no(0), nr(0), nc(0), fitness(0), isTested(false), randomSeed(1), id(0) {
	}

	CGPIndividual(int naToUse, int niToUse, int noToUse, int nrToUse, int ncToUse)
		: genome(NULL), na(naToUse), ni(niToUse), no(noToUse), nr(nrToUse), nc(ncToUse), fitness(0), isTested(false), randomSeed(1), id(0)
	{
		genome = new float[no+(nr*nc)*(2+na)];
		randomSeed = (int)genome;
	}

	~CGPIndividual() {
		if(genome) {
			delete [] genome;
			genome = NULL;
		}
	}

	void Setup(int naToUse, int niToUse, int noToUse, int nrToUse, int ncToUse) {
		na = naToUse;
		ni = niToUse;
		no = noToUse;
		nr = nrToUse;
		nc = ncToUse;
		genome = new float[no+(nr*nc)*(2+na)];
		randomSeed = (int)genome;
	}

	void CopyFrom(const CGPIndividual &other) const {
		memcpy(genome,other.genome,sizeof(float)*(no+(nr*nc)*(2+na)));
	}

	void SetupFrom(const CGPIndividual &other) {
		na = other.na;
		ni = other.ni;
		no = other.no;
		nr = other.nr;
		nc = other.nc;
		genome = new float[no+(nr*nc)*(2+na)];
		CopyFrom(other);
	}

	void SetOutputSource(int outputIndex, int sourceIndex) {
		genome[(nr*nc)*(2+na)+outputIndex] = sourceIndex;
	}

	void Randomize(RandomNumberSource &randomSource) {
		srand(randomSeed++);
		int r, c, a;
		// nr*nc*[function,parameter,input0,input1,...,inputNA-1], output0, output1,...,outputNO-1
		float *gp = genome;

		// Populate function genes
		for(c=0;c<nc;c++) { // for each column...
			const int nci = c*nr + ni;
			for(r=0;r<nr;r++) { // for each row in the current column...
				*gp++ = randomSource.NextInt(256); // Set the function to be 0-255
				*gp++ = (randomSource.Next() * 20) - 10; // Set the parameter to be -10 to +10
				for(a=0;a<na;a++) {
					*gp++ = randomSource.NextInt(nci); // Point to an input or anything in any prior column
				}
			}
		}

		// Populate output genes
		const int nrci = nr*nc + ni;
		for(int oi=0;oi<no;oi++) {
			*gp++ = randomSource.NextInt(nrci);
		}
	}
};

class CGPMutator {
public:
	float rate;
	int limit;
	RandomNumberSource &randomSource;

public:
	CGPMutator(int limitToUse, RandomNumberSource &randomSourceToUse)
		: limit(limitToUse), randomSource(randomSourceToUse) {
	}

	void Mutate(CGPIndividual *p, bool useHalfLimit=false) const {
		srand(p->randomSeed++);
		p->isTested = false;
		// nr*nc*[function,parameter,input0,input1,...,inputNA-1], output0, output1,...,outputNO-1
		const int geneCount = p->nc*p->nr*(2+p->na) + p->no;
		const int limitToUse = useHalfLimit? (limit/2)+1 : limit;
		for(int i=0;i<limitToUse;i++) {
			int index = randomSource.NextInt(geneCount);
			if(index < p->no) {
				// An output is to be mutated
				p->SetOutputSource(index,randomSource.NextInt(p->nc*p->nr + p->ni));
			} else {
				index -= p->no;
				int slot = index % (2+p->na);
				if(slot == 0) { // If changing the function
					// Mutate the function code to be 0-255.  The interpreter will % the function codes.
					p->genome[index] = randomSource.NextInt(256);
				} else if(slot == 1) { // If changing the parameter
					// Mutate the parameter to be -10 to +10
					p->genome[index] = (randomSource.Next() * 20) - 10;
				} else { // If changing a wiring source
					// Mutate the wiring source. This must come from a previous column or an input.
					int c = (index/(2+p->na)) / p->nr;
					int newConnection = randomSource.NextInt(p->ni + (c*p->nr)); 
					p->genome[index] = newConnection;
				}
			}
		}
	}
};

class CGPExecutor {
public:
	float *values;
	int startOfGenes;
	int startOfOutputs;
	int geneCount;
	int outputCount;

public:
	CGPExecutor()
		: values(NULL), startOfGenes(0), startOfOutputs(0), geneCount(0), outputCount(0) {
	}

	~CGPExecutor() {
		if(values != NULL) {
			delete [] values;
			values = NULL;
		}
	}

	void Setup(CGPIndividual *p) {
		geneCount = p->nr*p->nc;
		values = new float[p->ni + geneCount + p->no];
		startOfGenes = p->ni;
		startOfOutputs = startOfGenes + geneCount;
		outputCount = p->no;
	}

	void SetInput(int inputIndex, float value) const {
		values[inputIndex] = value;
	}

	void SetInputs(float *inputValues) const {
		float *vp = values;
		for(int i=0;i<startOfGenes;i++) {
			*vp++ = *inputValues++;
		}
	}

	float *GetOutputs() const {
		return values + startOfOutputs;
	}

	float GetOutput(int outputIndex) const {
		return values[startOfOutputs + outputIndex];
	}

	float *GetGenes() const {
		return values + startOfGenes;
	}

	void Execute(CGPIndividual *p) const {
		const float *gp = p->genome;
		const int geneSize = 2 + p->na;
		// Execute the genes
		float *vp = values + startOfGenes;
		for(int i=0;i<geneCount;i++,gp+=geneSize,vp++) {
			switch(((int)(*gp)) % 11) {
				case 0: // Constant
					*vp = gp[1];
					break;
				case 1: // Add
					*vp = values[(int)gp[2]] + values[(int)gp[3]];
					break;
				case 2: // Subtract
					*vp = values[(int)gp[2]] - values[(int)gp[3]];
					break;
				case 3: // Add parameter
					*vp = gp[1] + values[(int)gp[2]];
					break;
				case 4: // Subtract parameter
					*vp = values[(int)gp[2]] - gp[1];
					break;
				case 5: // Negate
					*vp = -values[(int)gp[2]];
					break;
				case 6: // Multiply
					*vp = values[(int)gp[2]] * values[(int)gp[3]];
					break;
				case 7: // Divide (guarded)
					if(values[(int)gp[3]] <= 1e-1 && values[(int)gp[3]] >= -1e-1) {
						*vp = values[(int)gp[2]];
					} else {
						*vp = values[(int)gp[2]] / values[(int)gp[3]];
					}
					break;
				case 8: // Modulus (guarded)
					if(values[(int)gp[3]] < 0) {
						*vp = -(int)values[(int)gp[2]];
					} else if(values[(int)gp[3]] < 1) {
						*vp = (int)values[(int)gp[2]];
					} else {
						*vp = (int)values[(int)gp[2]] % (int)values[(int)gp[3]];
					}
					break;
				case 9: // Average
					*vp = (values[(int)gp[2]] + values[(int)gp[3]]) / 2;
					break;
				case 10: // NOP
					*vp = values[(int)gp[2]];
					break;
			}
		}
		// Populate the outputs
		for(int i=0;i<outputCount;i++,vp++,gp++) {
			*vp = (int)values[(int)(*gp)] & 255;
		}
	}
};

class TestCases {
public:
	// howManyTests*[input0,input1,...,output0,output1,...,fitness0,fitness1,...]
	float *testData;
	int inputCount;
	int outputCount;
	int testSize;
	int howManyTests;
	int populationSize;

public:
	TestCases(int inputCountToUse, int outputCountToUse, int howManyTestsToUse, int populationSizeToUse)
			: testData(NULL), inputCount(inputCountToUse), outputCount(outputCountToUse), howManyTests(howManyTestsToUse), populationSize(populationSizeToUse) {
		testSize = inputCount + outputCount + populationSize;
		testData = new float[howManyTests * testSize];
	}

	~TestCases() {
		if(testData != NULL) {
			delete [] testData;
			testData = NULL;
		}
	}

	float *GetInputs(int testIndex) const {
		return testData + (testIndex * testSize);
	}

	float *GetOutputs(int testIndex) const {
		return testData + (testIndex * testSize) + inputCount;
	}

	void SetFitness(int testIndex, int individualIndex, float fitness) const {
		testData[testIndex * testSize + inputCount + outputCount + individualIndex] = fitness;
	}

	float GetFitness(int testIndex, int individualIndex) const {
		return testData[testIndex * testSize + inputCount + outputCount + individualIndex];
	}

	void CopyIndividualResults(int sourceIndex, int destIndex) const {
		for(int ti=0;ti<howManyTests;ti++) {
			SetFitness(ti,destIndex,GetFitness(ti,sourceIndex));
		}
	}

	float GetMaxFitness() const {
		float *fp = testData + inputCount + outputCount;
		float maxFitness = *fp;
		int pi;
		for(int ti=0;ti<howManyTests;ti++,fp+=inputCount+outputCount) {
			for(pi=0;pi<populationSize;pi++,fp++) {
				if(maxFitness < *fp) {
					maxFitness = *fp;
				}
			}
		}
		return maxFitness;
	}

	float GetMaxFitness(int individualIndex) const {
		float *fp = testData + inputCount + outputCount + individualIndex;
		float maxFitness = *fp;
		for(int ti=0;ti<howManyTests;ti++,fp+=testSize) {
			if(maxFitness < *fp) {
				maxFitness = *fp;
			}
		}
		return maxFitness;
	}
};

class CGPPopulation {
public:
	CGPIndividual *population;
	int populationSize;
	int na, ni, no, nr, nc;
	int nextID;

public:
	CGPPopulation(int populationSizeToUse, int naToUse, int niToUse, int noToUse, int nrToUse, int ncToUse)
		: population(NULL), populationSize(populationSizeToUse),
		na(naToUse), ni(niToUse), no(noToUse), nr(nrToUse), nc(ncToUse), nextID(0) {
		population = new CGPIndividual[populationSize];
		for(int i=0;i<populationSize;i++) {
			population[i].id = nextID+i;
			population[i].Setup(na,ni,no,nr,nc);
		}
		nextID = populationSize+1;
	}

	~CGPPopulation() {
		if(population) {
			delete [] population;
			population = NULL;
		}
	}

	void Randomize(RandomNumberSource &randomSource) {
		#pragma omp parallel for
		for(int pi=0;pi<populationSize;pi++) {
			population[pi].Randomize(randomSource);
		}
	}

	void Execute(TestCases &tests, bool testAllCases=true) const {
		#pragma omp parallel for
		for(int pi=0;pi<populationSize;pi++) {
			CGPIndividual *p = &population[pi];
			if(p->isTested) {
				continue;
			}
			CGPExecutor ex;
			ex.Setup(p);
			float fitness = 0, diff, t;
			for(int ti=0;ti<tests.howManyTests;ti++) {
				ex.SetInputs(tests.GetInputs(ti));
				ex.Execute(p);
				// Calculate fitness as sum of squared differences
				diff = 0;
				float *desiredOutput = tests.GetOutputs(ti);
				float *actualOutput = ex.GetOutputs();
				for(int oi=0;oi<p->no;oi++,desiredOutput++,actualOutput++) {
					t = (*desiredOutput - *actualOutput) * (*desiredOutput - *actualOutput);
					diff += t;
				}
				tests.SetFitness(ti,pi,diff);
				fitness += diff;
				// Stop if there is no chance of being better than or the same as the parent
				if(!testAllCases && fitness > population[0].fitness) {
					break;
				}
			}
			p->fitness = fitness;
			p->isTested = true;
		}
	}

	void PromoteBest(const CGPMutator &mutator, const TestCases &tests) {
		float bestFitness = population[0].fitness;
		int bestIndex = 0;
		for(int i=1;i<populationSize;i++) {
			if(population[i].fitness <= bestFitness) { // Note: "<=" enables neutral search
				bestFitness = population[i].fitness;
				bestIndex = i;
			}
		}
		if(bestIndex != 0) {
			population[0].CopyFrom(population[bestIndex]);
			population[0].fitness = population[bestIndex].fitness;
			population[0].isTested = true;
			population[0].id = population[bestIndex].id;
			tests.CopyIndividualResults(bestIndex,0);
		}
		#pragma omp parallel for
		for(int i=1;i<populationSize;i++) {
			population[i].CopyFrom(population[0]);
			population[i].id = nextID + i - 1;
			mutator.Mutate(&population[i]);
		}
		nextID += populationSize-1;
	}
};

};

#endif
