#include "DataVisFramework.h"
#include "../easyBMP/EasyBMP.h"

using namespace DataVisFramework;

static VisualizationCallback *callback = NULL;
static Stage *stages[3];
static float stagePositions[] = { 
	0, +1.25f, +2.5f,
	-1.5f, +1.25f, +2.5f,
	+1.5f, +1.25f, +2.5f
};

typedef unsigned int uint;

static GLSLProgram *floorProg = 0;

static int winWidth = 1280, winHeight = 720;
static int g_TotalErrors = 0;

// view params
static int ox, oy;
static int buttonState = 0;
static bool keyDown[256];

static vec3f cameraPos(0, -1, -4);
static vec3f cameraRot(0, 0, 0);
static vec3f cameraPosLag(cameraPos);
static vec3f cameraRotLag(cameraRot);

static vec3f lightPos(5.0, 5.0, -5.0);
static vec3f lightColor(1.0f, 1.0f, 0.8f);

const float inertia = 0.1f;
const float translateSpeed = 0.002f;
const float rotateSpeed = 0.2f;
const float walkSpeed = 0.05f;

enum { M_VIEW = 0 };
static int mode = M_VIEW;

// toggles
static bool paused = false;

static float modelView[16];

static GLuint floorTex = 0;

static void refreshFloorTexture();
static void createTexture(GLuint tex, GLenum target, GLint internalformat, GLenum format, int w, int h, void *data);
static GLuint createTexture(GLenum target, GLint internalformat, GLenum format, int w, int h, void *data);

// floor shader
static const char *floorVS = 
"varying vec4 vertexPosEye;  // vertex position in eye space  \n"
"varying vec3 normalEye;                                      \n"
"void main()                                                  \n"
"{                                                            \n"
"    gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;  \n"
"    gl_TexCoord[0] = gl_MultiTexCoord0;                      \n"
"    vertexPosEye = gl_ModelViewMatrix * gl_Vertex;           \n"
"    normalEye = gl_NormalMatrix * gl_Normal;                 \n"
"    gl_FrontColor = gl_Color;                                \n"
"}                                                            \n"
;

static const char *floorPS = 
"uniform sampler2D tex;                                                        \n"
"uniform vec3 lightPosEye; // light position in eye space                      \n"
"varying vec4 vertexPosEye;  // vertex position in eye space                   \n"
"varying vec3 normalEye;                                                       \n"
"void main()                                                                   \n"
"{                                                                             \n"
"    vec4 shadowPos = gl_TextureMatrix[0] * vertexPosEye;                      \n"
"    vec4 colorMap  = texture2D(tex, gl_TexCoord[0].xy);                       \n"
"    vec3 N = normalize(normalEye);                                            \n"
"    vec3 L = normalize(lightPosEye - vertexPosEye.xyz);                       \n"
"    float diffuse = max(0.50, dot(N, L));                                     \n"
"    gl_FragColor = vec4(gl_Color.xyz * colorMap.xyz * diffuse, 1.0);          \n"
"}                                                                             \n"
;

HeightMapVisualization::HeightMapVisualization(unsigned int nxToUse, unsigned int nyToUse)
	: Visualization(),
	nx(nxToUse), ny(nyToUse), vbo(0), vertices(NULL), heightScale(1.0),
	color(0,0,1),
	demoX(nxToUse/3), demoY(nyToUse/3), demoDX(+1), demoDY(+1),
	demoState(0)
{
	inverseHeightScale = (float)(1.0 / heightScale);
	vertices = new float[nx*ny*4];

	glGenBuffers(1, &vbo);
	glBindBuffer(GL_ARRAY_BUFFER, vbo);

	// initialize buffer object
	unsigned int size = nx * ny * 4 * sizeof(float);
	glBufferData(GL_ARRAY_BUFFER, size, 0, GL_DYNAMIC_DRAW);

	glBindBuffer(GL_ARRAY_BUFFER, 0);

	unsigned int ix, iy;
	float *vp = vertices;
	for(iy=0;iy<ny;iy++) {
		for(ix=0;ix<nx;ix++) {
			*vp++ = 1.0f * ((float)ix)/nx - 0.5f; // -0.5 to +0.5
			*vp++ = 1.0f * ((float)iy)/ny - 0.5f; // -0.5 to +0.5
			*vp++ = (float)(heightScale* Clip((float)(sin((float)ix/nx*2*3.14159) + sin((float)iy/ny*2*3.14159))/4));
			*vp++ = 1.0f;
		}
	}
	needsUpdate = true;
}

HeightMapVisualization::~HeightMapVisualization() {
	if(vbo) {
		//cudaGraphicsUnregisterResource(vbo_res);
		glBindBuffer(1, vbo);
		glDeleteBuffers(1, &vbo);
		vbo = 0;
	}
	if(vertices) {
		delete [] vertices;
	}
}

void HeightMapVisualization::ZeroHeights() {
	float *vp = vertices + 2;
	for(unsigned int count = nx * ny;count>0;count--,vp+=4) {
		*vp = 0;
	}
	needsUpdate = true;
}

void HeightMapVisualization::SetHeights(float *heights) {
	float *vp = vertices + 2;
	for(unsigned int count = nx * ny;count>0;count--,vp+=4) {
		*vp = Clip(*heights++) * heightScale;
	}
	needsUpdate = true;
}

void HeightMapVisualization::OnTranslateFromRawData() {
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	unsigned int size = nx * ny * 4 * sizeof(float);
	glBufferData(GL_ARRAY_BUFFER, size, vertices, GL_DYNAMIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
}

void HeightMapVisualization::OnRender() {
	// render from the vbo
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glVertexPointer(4, GL_FLOAT, 0, 0);

	glEnableClientState(GL_VERTEX_ARRAY);
	glColor3f(color[0], color[1], color[2]);
	glDrawArrays(GL_POINTS, 0, nx * ny);
	glDisableClientState(GL_VERTEX_ARRAY);
}

void HeightMapVisualization::AdvanceDemoAnimation() {
	demoState = (1+demoState) % (nx);
	int iy;
	unsigned int ix;
	float *vp;
	#pragma omp parallel for private (ix,vp)
	for(iy=0;iy<(int)ny;iy++) {
		vp = vertices + 4*iy*nx + 2;
		for(ix=0;ix<nx;ix++,vp+=4) {
			//SetHeight(ix,iy,(sin((float)(demoState+ix)/nx*2*3.14159) + sin((float)(demoState+iy)/ny*2*3.14159))/2);
			*vp = (float)(heightScale * Clip((float)(sin((float)(demoState+ix)/nx*2*3.14159) + sin((float)(demoState+iy)/ny*2*3.14159))/4));
		}
	}

	needsUpdate = true;
}

ImageVisualization::ImageVisualization(int widthToUse, int heightToUse)
		: width(widthToUse), height(heightToUse), rgba(NULL), texture(0) {
	int size = width*height*4;
	rgba = new unsigned char[size];
	#pragma omp parallel for
	for(int i=0;i<size;i++) {
		if((i & 3) == 3) {
			rgba[i] = 255; // set the alpha channel to no transparency
		} else {
			rgba[i] = 0;
		}
	}
	texture = createTexture(GL_TEXTURE_2D, GL_RGBA8, GL_RGBA, width, height, rgba);
	needsUpdate = true;
}

ImageVisualization::~ImageVisualization() {
	if(rgba != NULL) {
		delete [] rgba;
		rgba = NULL;
	}
}

void ImageVisualization::OnTranslateFromRawData() {
	createTexture(texture,GL_TEXTURE_2D,GL_RGBA8,GL_RGBA,width,height,rgba);
}

void ImageVisualization::OnRender() {
	floorProg->enable();
	floorProg->bindTexture("tex", texture, GL_TEXTURE_2D, 0);
	vec3f lightPosEye(0,10,5);
	floorProg->setUniformfv("lightPosEye", lightPosEye, 3);

	glColor3f(1.0, 1.0, 1.0);
	glNormal3f(0.0, 1.0, 0.0);
    glBegin(GL_QUADS);
    {
        float s = 1.f;
        float rep = 1.f;
        glTexCoord2f(0.f, 0.f); glVertex3f(-s/2, -s/2, 0);
        glTexCoord2f(rep, 0.f); glVertex3f(s/2, -s/2, 0);
        glTexCoord2f(rep, rep); glVertex3f(s/2, s/2, 0);
        glTexCoord2f(0.f, rep); glVertex3f(-s/2, s/2, 0);
	}
    glEnd();

    floorProg->disable();
}

void ImageVisualization::LoadBMP(const char *fileName) {
	BMP fileImage;
	if(!fileImage.ReadFromFile(fileName)) {
		return;
	}
	int fileWidth = fileImage.TellWidth();
	int fileHeight = fileImage.TellHeight();
	int widthToUse = width <= fileWidth? width : fileWidth;
	int heightToUse = height <= fileHeight? height : fileHeight;

	#pragma omp parallel for
	for(int y=0;y<heightToUse;y++) {
		unsigned char *row = GetPixel(0,y); // RGBA
		int fileImageY = heightToUse-1-y;
		for(int x=0;x<widthToUse;x++) {
			RGBApixel *p = fileImage(x,fileImageY);
			*row++ = p->Red;
			*row++ = p->Green;
			*row++ = p->Blue;
			*row++ = p->Alpha;
		}
	}

	needsUpdate = true;
}

void ImageVisualization::ApplyBayerFilter() {
	// Bayer images are 4-pixel squares with filtered color components:
	//   B G
	//   G R
	#pragma omp parallel for
	for(int y=0;y<height;y++) {
		unsigned char *row = GetPixel(0,y); // RGBA
		for(int x=0;x<width;x++,row+=4) {
			if(y & 1) { // if "G R" row
				if(x & 1) {
					// Keep Red only
					row[1] = row[2] = row[0]; // 0;
				} else {
					// Keep Green only
					row[0] = row[2] = row[1]; // 0;
				}
			} else { // if "B G" row
				if(x & 1) {
					// Keep Green only
					row[0] = row[2] = row[1]; // 0;
				} else {
					// Keep Blue only
					row[0] = row[1] = row[2]; // 0;
				}
			}
		}
	}

	needsUpdate = true;
}

void ImageVisualization::SetAll(unsigned char r, unsigned char g, unsigned char b, unsigned char a) {
	#pragma omp parallel for
	for(int y=0;y<height;y++) {
		unsigned char *row = GetPixel(0,y); // RGBA
		for(int x=0;x<width;x++) {
			*row++ = r;
			*row++ = g;
			*row++ = b;
			*row++ = a;
		}
	}

	needsUpdate = true;
}

static void cleanup() {
	for(int i=0;i<sizeof(stages)/sizeof(stages[0]);i++) {
		if(stages[i] != NULL) {
			delete stages[i];
			stages[i] = NULL;
		}
	}

	if(floorProg) {
		delete floorProg;
		floorProg = NULL;
	}

	if(floorTex)  {
		glDeleteTextures(1, &floorTex);
		floorTex = 0;
	}

	if(callback) {
		callback->OnQuit();
		callback = NULL;
	}
}

static void renderScene() {
    glEnable(GL_DEPTH_TEST);
    glDepthMask(GL_TRUE);

    // draw floor
	//refreshFloorTexture();
	floorProg->enable();
	floorProg->bindTexture("tex", floorTex, GL_TEXTURE_2D, 0);
	vec3f lightPosEye(0,10,5);
	floorProg->setUniformfv("lightPosEye", lightPosEye, 3);

	glColor3f(1.0, 1.0, 1.0);
	glNormal3f(0.0, 1.0, 0.0);
    glBegin(GL_QUADS);
    {
        float s = 20.f;
        float rep = 20.f;
        glTexCoord2f(0.f, 0.f); glVertex3f(-s, 0, -s);
        glTexCoord2f(rep, 0.f); glVertex3f(s, 0, -s);
        glTexCoord2f(rep, rep); glVertex3f(s, 0, s);
        glTexCoord2f(0.f, rep); glVertex3f(-s, 0, s);
    }
    glEnd();

    floorProg->disable();

	for(int i=0;i<sizeof(stages)/sizeof(stages[0]);i++) {
		if(stages[i] != NULL) {
			glPushMatrix();
			glTranslatef(stages[i]->center[0], stages[i]->center[1], stages[i]->center[2]);
			stages[i]->Render();
			glPopMatrix();
		}
	}

    glMatrixMode(GL_TEXTURE);
	glLoadIdentity();

	// draw light
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glTranslatef(lightPos.x, lightPos.y, lightPos.z);
    glColor3fv(&lightColor[0]);
    glutSolidSphere(0.1, 10, 5);
    glPopMatrix();
}

// main rendering loop
static void display()
{
    // move camera
    if (cameraPos[1] > 0.0f) cameraPos[1] = 0.0f;
    cameraPosLag += (cameraPos - cameraPosLag) * inertia;
    cameraRotLag += (cameraRot - cameraRotLag) * inertia;

    // view transform
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glRotatef(cameraRotLag[0], 1.0, 0.0, 0.0);
    glRotatef(cameraRotLag[1], 0.0, 1.0, 0.0);
    glTranslatef(cameraPosLag[0], cameraPosLag[1], cameraPosLag[2]);

    glGetFloatv(GL_MODELVIEW_MATRIX, modelView);

    // render
    glClearColor(0.0, 0.0, 0.0, 1.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);  
    renderScene();

    glutSwapBuffers();
    glutReportErrors();
}

// GLUT callback functions
static void reshape(int w, int h)
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(60.0, (float) w / (float) h, 0.01, 100.0);

    glMatrixMode(GL_MODELVIEW);
    glViewport(0, 0, w, h);
}

static void mouse(int button, int state, int x, int y)
{
    if (state == GLUT_DOWN) {
        buttonState |= 1<<button;
    } else if (state == GLUT_UP) {
        buttonState = 0;
	}
    int mods = glutGetModifiers();
    if (mods & GLUT_ACTIVE_SHIFT) {
        buttonState = 2;
    } else if (mods & GLUT_ACTIVE_CTRL) {
        buttonState = 3;
    }

    ox = x; oy = y;

    glutPostRedisplay();
}

// transfrom vector by matrix
static void xform(vec3f &v, vec3f &r, float *m)
{
  r.x = v.x*m[0] + v.y*m[4] + v.z*m[8] + m[12];
  r.y = v.x*m[1] + v.y*m[5] + v.z*m[9] + m[13];
  r.z = v.x*m[2] + v.y*m[6] + v.z*m[10] + m[14];
}

// transform vector by transpose of matrix (assuming orthonormal)
static void ixform(vec3f &v, vec3f &r, float *m)
{
  r.x = v.x*m[0] + v.y*m[1] + v.z*m[2];
  r.y = v.x*m[4] + v.y*m[5] + v.z*m[6];
  r.z = v.x*m[8] + v.y*m[9] + v.z*m[10];
}

static void ixformPoint(vec3f &v, vec3f &r, float *m)
{
    vec3f x;
    x.x = v.x - m[12];
    x.y = v.y - m[13];
    x.z = v.z - m[14];
    ixform(x, r, m);
}

static void dataVisMotion(int x, int y)
{
    float dx, dy;
    dx = (float)(x - ox);
    dy = (float)(y - oy);

    switch(mode) 
    {
	default:
    case M_VIEW:
        {
            if (buttonState == 1) {
                // left = rotate
                cameraRot[0] += dy * rotateSpeed;
                cameraRot[1] += dx * rotateSpeed;
            }
            if (buttonState == 2) {
                // middle = translate
                vec3f v = vec3f(dx*translateSpeed, -dy*translateSpeed, 0.0f);
                vec3f r;
                ixform(v, r, modelView);
                cameraPos += r;
            }
            if (buttonState == 3) {
                // left+middle = zoom
                vec3f v = vec3f(0.0, 0.0, dy*translateSpeed);
                vec3f r;
                ixform(v, r, modelView);
                cameraPos += r;
            } 
        }
        break;
    }

    ox = x; oy = y;
    glutPostRedisplay();
}

// commented out to remove unused parameter warnings in Linux
static void dataVisKey(unsigned char keyCode, int x, int y)
{
    switch (keyCode) 
    {
    case ' ':
		if(paused) {
			callback->OnResume();
		} else {
			callback->OnPause();
		}
        paused = !paused;
        break;
	case 'p':
	case 'P':
		if(!paused) {
			callback->OnPause();
			paused = true;
		}
		break;
	case 'r':
	case 'R':
		if(paused) {
			callback->OnResume();
			paused = false;
		}
		break;
    case 13:
        break;
    case '\033':
        cleanup();
        exit(0);
        break;
    case '1':
        break;
	default:
		callback->OnKey(keyCode);
		break;
    }

    keyDown[keyCode] = true;

    glutPostRedisplay();
}

static void keyUp(unsigned char key, int /*x*/, int /*y*/)
{
    keyDown[key] = false;
}

static void special(int k, int x, int y)
{
}

static void idle(void)
{
    // move camera in view direction
    /*
        0   4   8   12  x
        1   5   9   13  y
        2   6   10  14  z
    */
    if (keyDown['w']) {
        cameraPos[0] += modelView[2] * walkSpeed;
        cameraPos[1] += modelView[6] * walkSpeed;
        cameraPos[2] += modelView[10] * walkSpeed;
    }
    if (keyDown['s']) {
        cameraPos[0] -= modelView[2] * walkSpeed;
        cameraPos[1] -= modelView[6] * walkSpeed;
        cameraPos[2] -= modelView[10] * walkSpeed;
    }
    if (keyDown['a']) {
        cameraPos[0] += modelView[0] * walkSpeed;
        cameraPos[1] += modelView[4] * walkSpeed;
        cameraPos[2] += modelView[8] * walkSpeed;
    }
    if (keyDown['d']) {
        cameraPos[0] -= modelView[0] * walkSpeed;
        cameraPos[1] -= modelView[4] * walkSpeed;
        cameraPos[2] -= modelView[8] * walkSpeed;
    }
    if (keyDown['e']) {
        cameraPos[0] += modelView[1] * walkSpeed;
        cameraPos[1] += modelView[5] * walkSpeed;
        cameraPos[2] += modelView[9] * walkSpeed;
    }
    if (keyDown['q']) {
        cameraPos[0] -= modelView[1] * walkSpeed;
        cameraPos[1] -= modelView[5] * walkSpeed;
        cameraPos[2] -= modelView[9] * walkSpeed;
    }
	callback->OnIdle();

    glutPostRedisplay();
}

static void mainMenu(int i)
{
    dataVisKey((unsigned char) i, 0, 0);
}

static void initMenus()
{
    glutCreateMenu(mainMenu);
    glutAddMenuEntry("Reset view [1]", '1');
    glutAddMenuEntry("Pause [p]", 'p');
    glutAddMenuEntry("Resume [r]", 'r');
	glutAddMenuEntry("Quit (esc)", '\033');
    glutAttachMenu(GLUT_RIGHT_BUTTON);
}

static void createTexture(GLuint tex, GLenum target, GLint internalformat, GLenum format, int w, int h, void *data)
{
    glBindTexture(target, tex);
    glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(target, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glTexImage2D(target, 0, internalformat, w, h, 0, format, GL_UNSIGNED_BYTE, data);
}

static GLuint createTexture(GLenum target, GLint internalformat, GLenum format, int w, int h, void *data)
{
    GLuint tex;
    glGenTextures(1, &tex);
	createTexture(tex,target,internalformat,format,w,h,data);
	return tex;
	/*
    glBindTexture(target, tex);
    glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(target, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glTexImage2D(target, 0, internalformat, w, h, 0, format, GL_UNSIGNED_BYTE, data);
    return tex;
	*/
}

static GLuint loadTexture(char *filename)
{
    unsigned char *data = 0;
    unsigned int width, height;
    cutilCheckError( cutLoadPPM4ub(filename, &data, &width, &height));
    if (!data) {
        printf("Error opening file '%s'\n", filename);
        return 0;
    }
    printf("Loaded '%s', %d x %d pixels\n", filename, width, height);

    return createTexture(GL_TEXTURE_2D, GL_RGBA8, GL_RGBA, width, height, data);
}

static int floorWidth = 0, floorHeight = 0;
static unsigned char *floorData = NULL;

static GLuint loadFloorTexture(char *filename)
{
    unsigned char *data = 0;
    unsigned int width, height;
    cutilCheckError( cutLoadPPM4ub(filename, &data, &width, &height));
    if (!data) {
        printf("Error opening file '%s'\n", filename);
        return 0;
    }
    printf("Loaded '%s', %d x %d pixels\n", filename, width, height);

	floorData = data;
	floorWidth = width;
	floorHeight = height;

    return createTexture(GL_TEXTURE_2D, GL_RGBA8, GL_RGBA, width, height, data);
}

static void refreshFloorTexture() {
	int xOffset = rand() % floorWidth;
	int yOffset = rand() % floorHeight;
	unsigned char *p = floorData + (yOffset*floorWidth+xOffset)*4;
	*p++ = rand() & 255;
	*p++ = rand() & 255;
	*p++ = rand() & 255;

	createTexture(floorTex,GL_TEXTURE_2D,GL_RGBA8,GL_RGBA,floorWidth,floorHeight,floorData);
}

// initialize OpenGL
static void initGL(int *argc, char **argv)
{
    glutInit(argc, argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
    glutInitWindowSize(winWidth, winHeight);
    glutCreateWindow("DataVisFramework");

    glewInit();
    if (!glewIsSupported("GL_VERSION_2_0 GL_VERSION_1_5")) {
        fprintf(stderr, "The following required OpenGL extensions missing:\n\tGL_VERSION_2_0\n\tGL_VERSION_1_5\n");
        shrQAFinishExit(*argc, (const char **)argv, QA_WAIVED);
    }
    if (!glewIsSupported("GL_ARB_multitexture GL_ARB_vertex_buffer_object GL_EXT_geometry_shader4")) {
        fprintf(stderr, "The following required OpenGL extensions missing:\n\tGL_ARB_multitexture\n\tGL_ARB_vertex_buffer_object\n\tGL_EXT_geometry_shader4.\n");
        shrQAFinishExit(*argc, (const char **)argv, QA_WAIVED);
    }

#if defined (_WIN32)
    if (wglewIsSupported("WGL_EXT_swap_control")) {
        // disable vertical sync
        wglSwapIntervalEXT(0);
    }
#endif

    glEnable(GL_DEPTH_TEST);

    // load floor texture
    char* imagePath = cutFindFilePath("floortile.ppm", argv[0]);
    if (imagePath == NULL) {
        fprintf(stderr, "Error finding floor image file\n");
        shrQAFinishExit(*argc, (const char **)argv, QA_FAILED);
    }
    floorTex = loadFloorTexture(imagePath);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, 16.0f);

    floorProg = new GLSLProgram(floorVS, floorPS);	

    glutReportErrors();
}

void Stage::OnRender() {
	if(actor != NULL) {
		degrees += deltaDegrees;
		if(degrees < 0) {
			degrees += 360;
		} else if(degrees >= 360) {
			degrees -= 360;
		}
	    glRotatef(degrees, 0.0, 1.0, 0.0);
		actor->Render();
	}
}

class DefaultCallback : public VisualizationCallback {
public:
	// Nothing to do here, just accept all empty default functions.
};

VisualizationManager::VisualizationManager()
{
}

VisualizationManager::~VisualizationManager() {
}

void VisualizationManager::Setup(VisualizationCallback *callbackToUse, int *argc, char **argv) {
	if(callbackToUse == NULL) {
		callbackToUse = new DefaultCallback();
	}
	callback = callbackToUse;

	for(int i=0;i<sizeof(stages)/sizeof(stages[0]);i++) {
		stages[i] = NULL;
	}

    initGL( argc, argv );
}

void VisualizationManager::AddVisualization(Visualization *v) {
	for(int i=0;i<sizeof(stages)/sizeof(stages[0]);i++) {
		if(stages[i] == NULL) {
			vec3f center(stagePositions[i*3+0],stagePositions[i*3+1],stagePositions[i*3+2]);
			stages[i] = new Stage(center);
			stages[i]->actor = v;
			break;
		}
	}
}

void VisualizationManager::StartAndRunForever() {
    initMenus();

    glutDisplayFunc(display);
    glutReshapeFunc(reshape);
    glutMouseFunc(mouse);
    glutMotionFunc(dataVisMotion);
    glutKeyboardFunc(dataVisKey);
    glutKeyboardUpFunc(keyUp);
    glutSpecialFunc(special);
    glutIdleFunc(idle);

	atexit(cleanup);
    glutMainLoop();
}
