#ifndef DataVisFramework_H
#define DataVisFramework_H

#include <cstdlib>
#include <cstdio>
#include <algorithm>
#include <math.h>

#include <GL/glew.h>
#if defined(__APPLE__) || defined(__MACOSX)
#include <GLUT/glut.h>
#else
#include <GL/freeglut.h>
#endif
#if defined (_WIN32)
#include <GL/wglew.h>
#endif

#include <shrQATest.h>
#include <cutil_inline.h>    // includes cuda.h and cuda_runtime_api.h
#include <cutil_gl_inline.h>
#include <cuda_gl_interop.h>

#include <nvmath.h>

#include "paramgl.h"
#include "GLSLProgram.h"

#include <iostream>
#include <fstream>
using namespace nv;

namespace DataVisFramework {

// Visualizations draw themselves into a 1x1x1 cube centered at <0,0,0>.
class Visualization {
public:
	bool paused;
	volatile bool needsUpdate;

public:
	Visualization()
		: paused(false), needsUpdate(false)
	{
	}

	virtual ~Visualization() {
	}

	void Render() {
		if(needsUpdate) {
			needsUpdate = false;
			OnTranslateFromRawData();
		}
		OnRender();
	}

	virtual void OnRender() = 0;

	virtual void OnPause() {
		paused = true;
	}

	virtual void OnResume() {
		paused = false;
	}

	virtual void OnTranslateFromRawData() {
		// Nothing to do here by default
	}
};

class Stage : public Visualization {
public:
	vec3f center;
	float degrees;
	float deltaDegrees;
	Visualization *actor;

public:
	Stage(vec3f centerToUse)
		: center(centerToUse), degrees(0), deltaDegrees(0.2f), actor(NULL)
	{
	}

	virtual ~Stage() {
	}

	virtual void OnRender();

	virtual void OnPause() {
		if(actor != NULL) {
			actor->OnPause();
		}
	}

	virtual void OnResume() {
		if(actor != NULL) {
			actor->OnResume();
		}
	}
};

class HeightMapVisualization : public Visualization {
public:
	unsigned int nx, ny;
	GLuint vbo;
	float *vertices;
	float heightScale, inverseHeightScale;
	vec3f color;

	int demoX, demoY, demoDX, demoDY;
	int demoState;

public:
	HeightMapVisualization(unsigned int nxToUse, unsigned int nyToUse);
	virtual ~HeightMapVisualization();

	void SetHeight(unsigned int x, unsigned int y, float h) {
		vertices[(x + y*nx)*4 + 2] = heightScale * Clip(h);
		needsUpdate = true;
	}

	float GetHeight(unsigned int x, unsigned int y) const {
		return vertices[(x + y*nx)*4 + 2] * inverseHeightScale;
	}

	void ZeroHeights();
	void SetHeights(float *heights);
	virtual void OnTranslateFromRawData();
	virtual void OnRender();
	void AdvanceDemoAnimation();

	static float Clip(float h) {
		return h < -0.5f? -0.5f : h > 0.5f? 0.5f : h;
	}
};

class ImageVisualization : public Visualization {
public:
	int width;
	int height;
	unsigned char *rgba;
	GLuint texture;

public:
	ImageVisualization(int widthToUse, int heightToUse);
	virtual ~ImageVisualization();

	virtual void OnTranslateFromRawData();
	virtual void OnRender();

	void SetAll(unsigned char r, unsigned char g, unsigned char b, unsigned char a=255);

	void SetPixel(int x, int y, unsigned char r, unsigned char g, unsigned char b, unsigned char a=255) {
		unsigned char *p = rgba + ((x + y*width)<<2);
		p[0] = r;
		p[1] = g;
		p[2] = b;
		p[3] = a;
		needsUpdate = true;
	}

	unsigned char *GetPixel(int x, int y) const {
		return rgba + ((x + y*width)<<2);
	}

	void LoadBMP(const char *fileName);

	void ApplyBayerFilter();
};

class VisualizationCallback {
public:
	virtual void OnPause() {
		// Nothing to do by default
	}

	virtual void OnResume() {
		// Nothing to do by default
	}

	virtual void OnQuit() {
		// Nothing to do by default
	}

	virtual void OnKey(unsigned char keyCode) {
		// Nothing to do by default
	}

	virtual void OnIdle() {
		// Nothing to do by default
	}
};

class VisualizationManager {
public:
	VisualizationManager();
	virtual ~VisualizationManager();

	virtual void Setup(VisualizationCallback *callbackToUse, int *argc, char **argv);
	virtual void AddVisualization(Visualization *v);
	virtual void StartAndRunForever();
};

};

#endif
