#include "DataVisFramework.h"
#include "CGP.h"
using namespace DataVisFramework;
using namespace CartesianGeneticProgramming;

class DeBayerSimulation : public VisualizationCallback {
public:
	bool isSetup;
	bool isPaused;
	HeightMapVisualization *hm;
	RandomNumberSource randomSource;
	CGPMutator *mutator;
	TestCases *tests;
	CGPPopulation *population;
	int ni, no, nr, nc, populationSize, howManyTests;

	int splotWidth, splotHeight;
	int generationNumber;
	int frameNumber;
	int priorBestID;
	float priorBestFitness;
	int changeCount;

	vec3f colors[10];

	ImageVisualization *rawImage;
	ImageVisualization *processedImage;
	int imageTestIndex;

public:
	DeBayerSimulation()
			: isSetup(false), isPaused(false), hm(NULL), mutator(NULL), tests(NULL), population(NULL),
			ni(4), no(3), nr(8), nc(8), populationSize(5), howManyTests(200),
			splotWidth(4), splotHeight(20), generationNumber(0), frameNumber(0),
			priorBestID(-1), priorBestFitness(-1), changeCount(0),
			rawImage(NULL), processedImage(NULL), imageTestIndex(0) {
		srand(1);

		for(int i=0;i<sizeof(colors)/sizeof(colors[0]);i++) {
			colors[i][0] = ((rand() % 128) + 127) / 255.;
			colors[i][1] = ((rand() % 128) + 127) / 255.;
			colors[i][2] = ((rand() % 128) + 127) / 255.;
		}
	}

	~DeBayerSimulation() {
		OnQuit();
	}

	void Setup() {
		isSetup = true;

		hm = new HeightMapVisualization(splotWidth*howManyTests,splotHeight*populationSize);
		hm->color[0] = 0; hm->color[1] = 1; hm->color[2] = 0;

		rawImage = new ImageVisualization(640,480);
		rawImage->LoadBMP("data/pebbles_640x480_rgb.bmp");
		rawImage->ApplyBayerFilter();

		processedImage = new ImageVisualization(320,240);
		processedImage->LoadBMP("data/pebbles_320x240_rgb.bmp");

		mutator = new CGPMutator((int)((nr*nc*(2+2)+no)*0.08),randomSource);
		tests = new TestCases(ni,no,howManyTests,populationSize);
		population = new CGPPopulation(populationSize,2,ni,no,nr,nc);
		population->Randomize(randomSource);
		LoadTestCases();

		processedImage->SetAll(128,128,128,255);
	}

	int R(float x, float y) const {
		return (int)((sin(x/(howManyTests*2) * 2 * 3.14159)+1) / 2 * 255);
	}

	int G(float x, float y) const {
		return (int)((cos(x/(howManyTests*2) * 2 * 3.14159)+1) / 2 * 255);
	}

	int B(float x, float y) const {
		return (int)((cos((y+x)/(howManyTests*2+2) * 2 * 3.14159)+1) / 2 * 255);
	}

	void LoadTestCases() {
		// Setup fixed test cases
		int ti;
		const int fixedCount = 10;
		for(ti=0;ti<fixedCount;ti++) {
			unsigned char gray = (unsigned char)(ti * (255.f/(fixedCount-1)));

			float *inputs = tests->GetInputs(ti);
			inputs[0] = gray;
			inputs[1] = gray;
			inputs[2] = gray;
			inputs[3] = gray;

			float *outputs = tests->GetOutputs(ti);
			outputs[0] = gray;
			outputs[1] = gray;
			outputs[2] = gray;
		}

		// Setup random test cases
		for(;ti<howManyTests;ti++) {
			int processedX = rand() % processedImage->width;
			int processedY = rand() % processedImage->height;
			int rawX = processedX * 2;
			int rawY = processedY * 2;

			float *inputs = tests->GetInputs(ti);
			inputs[0] = rawImage->GetPixel(rawX+0,rawY+0)[2]; // Blue
			inputs[1] = rawImage->GetPixel(rawX+1,rawY+0)[1]; // Green
			inputs[2] = rawImage->GetPixel(rawX+0,rawY+1)[1]; // Green
			inputs[3] = rawImage->GetPixel(rawX+1,rawY+1)[0]; // Red

			float *outputs = tests->GetOutputs(ti);
			unsigned char *processedRGBA = processedImage->GetPixel(processedX,processedY);
			outputs[0] = processedRGBA[0];
			outputs[1] = processedRGBA[1];
			outputs[2] = processedRGBA[2];
		}
	}

	void Visualize() {
		if(frameNumber % 100 == 0) {
			if(frameNumber == 0 || priorBestFitness != population->population[0].fitness) {
				priorBestID = population->population[0].id;
				priorBestFitness = population->population[0].fitness;
				changeCount++;
				hm->color = colors[changeCount % (sizeof(colors)/sizeof(colors[0]))];
				ExecuteToImage();
			}
			const float maxFitness = tests->GetMaxFitness(0);
			const float scale = maxFitness >= 1? 1/maxFitness : 1;
			#pragma omp parallel for
			for(int pi=0;pi<populationSize;pi++) {
				for(int ti=0;ti<howManyTests;ti++) {
					const float h = tests->GetFitness(ti,pi) * scale;
					const int xBase = splotWidth * ti;
					const int xMax = xBase + splotWidth;
					int y = splotHeight * pi;
					const int yMax = y + splotHeight;
					int x;
					for(;y<yMax;y++) {
						for(x=xBase;x<xMax;x++) {
							hm->SetHeight(x,y,h);
						}
					}
				}
			}
		}
		frameNumber++;
	}

	void ExecuteToImage() {
		processedImage->SetAll(0,0,0,255);

		CGPIndividual *p = &population->population[0];
		CGPExecutor ex;
		ex.Setup(p);
		for(int y=0;y<processedImage->height;y++) {
			int rawY = y*2;
			for(int x=0;x<processedImage->width;x++) {
				int rawX = x*2;
				// Inputs are B G | G B from the two rows of the Bayer image
				ex.SetInput(0,rawImage->GetPixel(rawX+0,rawY+0)[2]);
				ex.SetInput(1,rawImage->GetPixel(rawX+1,rawY+0)[1]);
				ex.SetInput(2,rawImage->GetPixel(rawX+0,rawY+1)[1]);
				ex.SetInput(3,rawImage->GetPixel(rawX+1,rawY+1)[0]);
				ex.Execute(p);
				float *actualOutput = ex.GetOutputs();
				unsigned char *rgba = processedImage->GetPixel(x,y);
				rgba[0] = (unsigned char)actualOutput[0];
				rgba[1] = (unsigned char)actualOutput[1];
				rgba[2] = (unsigned char)actualOutput[2];
			}
		}
		processedImage->needsUpdate = true;
	}

	virtual void OnPause() {
		isPaused = true;
	}

	virtual void OnResume() {
		isPaused = false;
	}

	virtual void OnQuit() {
		if(rawImage) {
			delete rawImage;
			rawImage = NULL;
		}
		if(processedImage) {
			delete processedImage;
			processedImage = NULL;
		}
		if(hm) {
			delete hm;
			hm = NULL;
		}
		if(mutator) {
			delete mutator;
			mutator = NULL;
		}
		if(tests) {
			delete tests;
			tests = NULL;
		}
		if(population) {
			delete population;
			population = NULL;
		}
	}

	virtual void OnKey(unsigned char keyCode) {
		// Nothing to do by default
	}

	virtual void OnIdle() {
		if(isSetup && !isPaused) {
			for(int i=0;i<10;i++) {
				population->Execute(*tests,generationNumber<5);
				population->PromoteBest(*mutator,*tests);
				generationNumber++;
				if(generationNumber % 250 == 0) {
					printf("@%d: %g, %g, %g, %g, %g\n",generationNumber,
							population->population[0].fitness,
							population->population[1].fitness,
							population->population[2].fitness,
							population->population[3].fitness,
							population->population[4].fitness);
				}
			}
			Visualize();
		}
	}
};

////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char** argv) {
	DeBayerSimulation s;

	VisualizationManager vm;
	vm.Setup(&s,&argc,argv);
	s.Setup();
	vm.AddVisualization(s.processedImage);
	vm.AddVisualization(s.hm);
	vm.AddVisualization(s.rawImage);

    if(cutCheckCmdLineFlag(argc, (const char **)argv, "device")) {
      cutilGLDeviceInit(argc, argv);
    } else {
        cudaGLSetGLDevice(cutGetMaxGflopsDeviceId());
    }

	vm.StartAndRunForever();

    cutilDeviceReset();
}
